/*
	$Id: Pool.c 117 2006-03-07 11:53:10Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <string.h>
#include "impl/InputStream.h"
#include "impl/OutputStream.h"
#include "impl/Pool.h"
#include "Pool.h"
#include "LinkedList.h"

struct Bucket {
    size_t size;
    unsigned char data[0];
};

union ListElement {
    struct Bucket bucket;
};

static struct Bucket *
createBucket(const void *data, size_t size)
{
    struct Bucket *bucket;

    if ((bucket = malloc(sizeof(*bucket) + size)) == NULL) {
	return NULL;
    }
    bucket->size = size;
    memcpy(bucket->data, data, size);
    return bucket;
}

static void
freeBucket(struct Bucket *bucket)
{
    free(bucket);
}

union Pool
Pool_create(void)
{
    struct impl_Pool *impl;

    if ((impl = malloc(sizeof(struct impl_Pool))) == NULL) {
	return (union Pool)null;
    }
    if ((impl->bucket = LinkedList_create()).super.object == NULL) {
	free(impl);
	return (union Pool)null;
    }
    impl->size = 0;
    return (union Pool)impl;
}

size_t
Pool_getSize(union Pool p)
{
    struct impl_Pool *impl = p.impl;

    return impl->size;
}

void
Pool_free(union Pool p)
{
    struct impl_Pool *impl = p.impl;
    union ListElement *e;

    while ((e = LinkedList_getFirst(impl->bucket)) != NULL)
	freeBucket(&e->bucket);
    LinkedList_free(impl->bucket);
    free(impl);
}

/* __final__ */ struct impl_PoolOutputStream {
    union {
        struct impl_OutputStream outputStream;
    } super;
    struct impl_Pool *pool;
};

union POS {
    struct impl_OutputStream *outputStream;
    struct impl_PoolOutputStream *impl;
} TRANSPARENT_UNION;

static ssize_t
POS_write(union POS this, const void *data, size_t size)
{
    struct impl_PoolOutputStream *impl = this.impl;
    struct impl_Pool *pool = impl->pool;
    union {
	union ListElement *listElement;
	struct Bucket *bucket;
    } e;

    if ((e.bucket = createBucket(data, size)) == NULL) {
	return -1;
    }
    if (LinkedList_append(pool->bucket, e.listElement) < 0) {
	freeBucket(e.bucket);
	return -1;
    }
    pool->size += size;
    return size;
}

static int
POS_close(union POS this)
{
    struct impl_PoolOutputStream *impl = this.impl;

    free(impl);
    return 0;
}

union OutputStream
Pool_getOutputStream(union Pool p)
{
    struct impl_PoolOutputStream *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
        return (union OutputStream)null;
    }
    impl_OutputStream_new(&impl->super.outputStream,
                          POS_write,
                          POS_close);
    impl->pool = p.impl;
    return (union OutputStream)&impl->super.outputStream;
}

/* __final__ */ struct impl_PoolInputStream {
    union {
        struct impl_InputStream inputStream;
    } super;
    struct impl_Pool *pool;
    struct Bucket *bucket;
    size_t offset;
};

union PIS {
    struct impl_InputStream *inputStream;
    struct impl_PoolInputStream *impl;
} TRANSPARENT_UNION;

static ssize_t
PIS_read(union PIS this, void *data, size_t size)
{
    struct impl_PoolInputStream *impl = this.impl;
    struct impl_Pool *pool = impl->pool;
    size_t len;
    union ListElement *e;

    if (impl->bucket == NULL) {
	if ((e = LinkedList_getFirst(pool->bucket)) == NULL)
	    return 0;
	impl->bucket = &e->bucket;
	impl->offset = 0;
    }

    if ((len = impl->bucket->size - impl->offset) > size) {
	len = size;
    }
    memcpy(data, impl->bucket->data + impl->offset, len);
    impl->offset += len;

    if (impl->offset == impl->bucket->size) {
	freeBucket(impl->bucket);
	impl->bucket = NULL;
    }
    return len;
}

static int
PIS_close(union PIS this)
{
    struct impl_PoolInputStream *impl = this.impl;

    free(impl);
    return 0;
}

union InputStream
Pool_getInputStream(union Pool p)
{
    struct impl_PoolInputStream *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
        return (union InputStream)null;
    }
    impl->super.inputStream.read = PIS_read;
    impl->super.inputStream.close = PIS_close;
    impl->pool = p.impl;
    impl->bucket = NULL;
    impl->offset = 0;
    return (union InputStream)&impl->super.inputStream;
}
